from eth_typing import HexAddress as HexAddress
from eth_typing.abi import Decodable as Decodable, TypeStr as TypeStr
from faster_eth_abi._codec import decode_c as decode_c, encode_c as encode_c
from faster_eth_abi.decoding import ContextFramesBytesIO as ContextFramesBytesIO
from faster_eth_abi.exceptions import (
    EncodingError as EncodingError,
    MultipleEntriesFound as MultipleEntriesFound,
)
from faster_eth_abi.registry import ABIRegistry as ABIRegistry
from faster_eth_abi.typing import (
    AddressTypeStr as AddressTypeStr,
    BoolTypeStr as BoolTypeStr,
    BytesTypeStr as BytesTypeStr,
    IntTypeStr as IntTypeStr,
    StringTypeStr as StringTypeStr,
    UintTypeStr as UintTypeStr,
)
from typing import Any, Iterable, overload

DecodesToIntTypeStr = UintTypeStr | IntTypeStr

class BaseABICoder:
    """
    Base class for porcelain coding APIs.  These are classes which wrap
    instances of :class:`~faster_eth_abi.registry.ABIRegistry` to provide last-mile
    coding functionality.
    """

    def __init__(self, registry: ABIRegistry) -> None:
        """
        Constructor.

        :param registry: The registry providing the encoders to be used when
            encoding values.
        """

class ABIEncoder(BaseABICoder):
    """
    Wraps a registry to provide last-mile encoding functionality.
    """

    def encode(self, types: Iterable[TypeStr], args: Iterable[Any]) -> bytes:
        """
        Encodes the python values in ``args`` as a sequence of binary values of
        the ABI types in ``types`` via the head-tail mechanism.

        :param types: A list or tuple of string representations of the ABI types
            that will be used for encoding e.g.  ``('uint256', 'bytes[]',
            '(int,int)')``
        :param args: A list or tuple of python values to be encoded.

        :returns: The head-tail encoded binary representation of the python
            values in ``args`` as values of the ABI types in ``types``.
        """

    def is_encodable(self, typ: TypeStr, arg: Any) -> bool:
        """
        Determines if the python value ``arg`` is encodable as a value of the
        ABI type ``typ``.

        :param typ: A string representation for the ABI type against which the
            python value ``arg`` will be checked e.g. ``'uint256'``,
            ``'bytes[]'``, ``'(int,int)'``, etc.
        :param arg: The python value whose encodability should be checked.

        :returns: ``True`` if ``arg`` is encodable as a value of the ABI type
            ``typ``.  Otherwise, ``False``.
        """

    def is_encodable_type(self, typ: TypeStr) -> bool:
        """
        Returns ``True`` if values for the ABI type ``typ`` can be encoded by
        this codec.

        :param typ: A string representation for the ABI type that will be
            checked for encodability e.g. ``'uint256'``, ``'bytes[]'``,
            ``'(int,int)'``, etc.

        :returns: ``True`` if values for ``typ`` can be encoded by this codec.
            Otherwise, ``False``.
        """

class ABIDecoder(BaseABICoder):
    """
    Wraps a registry to provide last-mile decoding functionality.
    """

    stream_class = ContextFramesBytesIO
    @overload
    def decode(
        self, types: tuple[AddressTypeStr], data: Decodable, strict: bool = True
    ) -> tuple[HexAddress]: ...
    @overload
    def decode(
        self, types: tuple[BytesTypeStr], data: Decodable, strict: bool = True
    ) -> tuple[bytes]: ...
    @overload
    def decode(
        self, types: tuple[StringTypeStr], data: Decodable, strict: bool = True
    ) -> tuple[str]: ...
    @overload
    def decode(
        self, types: tuple[DecodesToIntTypeStr], data: Decodable, strict: bool = True
    ) -> tuple[int]: ...
    @overload
    def decode(
        self, types: tuple[BoolTypeStr], data: Decodable, strict: bool = True
    ) -> tuple[bool]: ...
    @overload
    def decode(
        self, types: tuple[TypeStr], data: Decodable, strict: bool = True
    ) -> tuple[Any]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, str]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, int]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, str]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, int]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, bool]: ...
    @overload
    def decode(
        self, types: tuple[BytesTypeStr, TypeStr], data: Decodable, strict: bool = True
    ) -> tuple[bytes, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, str]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, int]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, bool]: ...
    @overload
    def decode(
        self, types: tuple[StringTypeStr, TypeStr], data: Decodable, strict: bool = True
    ) -> tuple[str, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, str]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, int]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, str]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, int]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, bool]: ...
    @overload
    def decode(
        self, types: tuple[BoolTypeStr, TypeStr], data: Decodable, strict: bool = True
    ) -> tuple[bool, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, HexAddress]: ...
    @overload
    def decode(
        self, types: tuple[TypeStr, BytesTypeStr], data: Decodable, strict: bool = True
    ) -> tuple[Any, bytes]: ...
    @overload
    def decode(
        self, types: tuple[TypeStr, StringTypeStr], data: Decodable, strict: bool = True
    ) -> tuple[Any, str]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, int]: ...
    @overload
    def decode(
        self, types: tuple[TypeStr, BoolTypeStr], data: Decodable, strict: bool = True
    ) -> tuple[Any, bool]: ...
    @overload
    def decode(
        self, types: tuple[TypeStr, TypeStr], data: Decodable, strict: bool = True
    ) -> tuple[Any, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, AddressTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, HexAddress, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, AddressTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, HexAddress, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, AddressTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, HexAddress, str]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, AddressTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, HexAddress, int]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, AddressTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, HexAddress, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, AddressTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, HexAddress, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, BytesTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, bytes, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, BytesTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, bytes, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, BytesTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, bytes, str]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, BytesTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, bytes, int]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, BytesTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, bytes, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, BytesTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, bytes, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, StringTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, str, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, StringTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, str, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, StringTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, str, str]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, StringTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, str, int]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, StringTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, str, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, StringTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, str, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, DecodesToIntTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, int, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, DecodesToIntTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, int, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, DecodesToIntTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, int, str]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, DecodesToIntTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, int, int]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, DecodesToIntTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, int, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, DecodesToIntTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, int, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, BoolTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, bool, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, BoolTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, bool, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, BoolTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, bool, str]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, BoolTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, bool, int]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, BoolTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, bool, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, BoolTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, bool, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, TypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, Any, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, TypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, Any, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, TypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, Any, str]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, TypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, Any, int]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, TypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, Any, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[AddressTypeStr, TypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress, Any, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, AddressTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, HexAddress, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, AddressTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, HexAddress, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, AddressTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, HexAddress, str]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, AddressTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, HexAddress, int]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, AddressTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, HexAddress, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, AddressTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, HexAddress, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, BytesTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, bytes, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, BytesTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, bytes, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, BytesTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, bytes, str]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, BytesTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, bytes, int]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, BytesTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, bytes, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, BytesTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, bytes, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, StringTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, str, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, StringTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, str, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, StringTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, str, str]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, StringTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, str, int]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, StringTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, str, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, StringTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, str, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, DecodesToIntTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, int, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, DecodesToIntTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, int, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, DecodesToIntTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, int, str]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, DecodesToIntTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, int, int]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, DecodesToIntTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, int, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, DecodesToIntTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, int, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, BoolTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, bool, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, BoolTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, bool, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, BoolTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, bool, str]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, BoolTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, bool, int]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, BoolTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, bool, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, BoolTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, bool, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, TypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, Any, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, TypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, Any, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, TypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, Any, str]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, TypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, Any, int]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, TypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, Any, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[BytesTypeStr, TypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes, Any, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, AddressTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, HexAddress, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, AddressTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, HexAddress, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, AddressTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, HexAddress, str]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, AddressTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, HexAddress, int]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, AddressTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, HexAddress, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, AddressTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, HexAddress, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, BytesTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, bytes, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, BytesTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, bytes, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, BytesTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, bytes, str]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, BytesTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, bytes, int]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, BytesTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, bytes, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, BytesTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, bytes, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, StringTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, str, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, StringTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, str, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, StringTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, str, str]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, StringTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, str, int]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, StringTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, str, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, StringTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, str, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, DecodesToIntTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, int, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, DecodesToIntTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, int, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, DecodesToIntTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, int, str]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, DecodesToIntTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, int, int]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, DecodesToIntTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, int, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, DecodesToIntTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, int, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, BoolTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, bool, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, BoolTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, bool, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, BoolTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, bool, str]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, BoolTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, bool, int]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, BoolTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, bool, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, BoolTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, bool, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, TypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, Any, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, TypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, Any, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, TypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, Any, str]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, TypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, Any, int]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, TypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, Any, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[StringTypeStr, TypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str, Any, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, AddressTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, HexAddress, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, AddressTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, HexAddress, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, AddressTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, HexAddress, str]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, AddressTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, HexAddress, int]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, AddressTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, HexAddress, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, AddressTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, HexAddress, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, BytesTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, bytes, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, BytesTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, bytes, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, BytesTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, bytes, str]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, BytesTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, bytes, int]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, BytesTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, bytes, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, BytesTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, bytes, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, StringTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, str, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, StringTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, str, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, StringTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, str, str]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, StringTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, str, int]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, StringTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, str, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, StringTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, str, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, DecodesToIntTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, int, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, DecodesToIntTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, int, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, DecodesToIntTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, int, str]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, DecodesToIntTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, int, int]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, DecodesToIntTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, int, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, DecodesToIntTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, int, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, BoolTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, bool, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, BoolTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, bool, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, BoolTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, bool, str]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, BoolTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, bool, int]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, BoolTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, bool, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, BoolTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, bool, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, TypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, Any, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, TypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, Any, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, TypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, Any, str]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, TypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, Any, int]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, TypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, Any, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[DecodesToIntTypeStr, TypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int, Any, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, AddressTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, HexAddress, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, AddressTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, HexAddress, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, AddressTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, HexAddress, str]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, AddressTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, HexAddress, int]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, AddressTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, HexAddress, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, AddressTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, HexAddress, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, BytesTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, bytes, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, BytesTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, bytes, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, BytesTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, bytes, str]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, BytesTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, bytes, int]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, BytesTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, bytes, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, BytesTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, bytes, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, StringTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, str, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, StringTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, str, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, StringTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, str, str]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, StringTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, str, int]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, StringTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, str, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, StringTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, str, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, DecodesToIntTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, int, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, DecodesToIntTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, int, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, DecodesToIntTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, int, str]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, DecodesToIntTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, int, int]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, DecodesToIntTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, int, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, DecodesToIntTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, int, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, BoolTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, bool, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, BoolTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, bool, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, BoolTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, bool, str]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, BoolTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, bool, int]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, BoolTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, bool, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, BoolTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, bool, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, TypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, Any, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, TypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, Any, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, TypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, Any, str]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, TypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, Any, int]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, TypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, Any, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[BoolTypeStr, TypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bool, Any, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, AddressTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, HexAddress, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, AddressTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, HexAddress, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, AddressTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, HexAddress, str]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, AddressTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, HexAddress, int]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, AddressTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, HexAddress, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, AddressTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, HexAddress, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, BytesTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, bytes, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, BytesTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, bytes, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, BytesTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, bytes, str]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, BytesTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, bytes, int]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, BytesTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, bytes, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, BytesTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, bytes, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, StringTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, str, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, StringTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, str, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, StringTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, str, str]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, StringTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, str, int]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, StringTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, str, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, StringTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, str, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, DecodesToIntTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, int, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, DecodesToIntTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, int, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, DecodesToIntTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, int, str]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, DecodesToIntTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, int, int]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, DecodesToIntTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, int, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, DecodesToIntTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, int, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, BoolTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, bool, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, BoolTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, bool, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, BoolTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, bool, str]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, BoolTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, bool, int]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, BoolTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, bool, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, BoolTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, bool, Any]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, TypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, Any, HexAddress]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, TypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, Any, bytes]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, TypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, Any, str]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, TypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, Any, int]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, TypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, Any, bool]: ...
    @overload
    def decode(
        self,
        types: tuple[TypeStr, TypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[Any, Any, Any]: ...
    @overload
    def decode(
        self, types: Iterable[AddressTypeStr], data: Decodable, strict: bool = True
    ) -> tuple[HexAddress, ...]: ...
    @overload
    def decode(
        self, types: Iterable[BytesTypeStr], data: Decodable, strict: bool = True
    ) -> tuple[bytes, ...]: ...
    @overload
    def decode(
        self, types: Iterable[StringTypeStr], data: Decodable, strict: bool = True
    ) -> tuple[str, ...]: ...
    @overload
    def decode(
        self, types: Iterable[DecodesToIntTypeStr], data: Decodable, strict: bool = True
    ) -> tuple[int, ...]: ...
    @overload
    def decode(
        self, types: Iterable[BoolTypeStr], data: Decodable, strict: bool = True
    ) -> tuple[bool, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[AddressTypeStr | BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress | bytes, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[AddressTypeStr | StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress | str, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[AddressTypeStr | DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress | int, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[AddressTypeStr | BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress | bool, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[BytesTypeStr | StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes | str, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[BytesTypeStr | DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes | int, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[BytesTypeStr | BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes | bool, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[StringTypeStr | DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str | int, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[StringTypeStr | BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str | bool, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[DecodesToIntTypeStr | BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[int | bool, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[AddressTypeStr | BytesTypeStr | StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress | bytes | str, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[AddressTypeStr | BytesTypeStr | DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress | bytes | int, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[AddressTypeStr | BytesTypeStr | BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress | bytes | bool, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[AddressTypeStr | StringTypeStr | DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress | str | int, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[AddressTypeStr | StringTypeStr | BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress | str | bool, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[AddressTypeStr | DecodesToIntTypeStr | BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress | int | bool, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[BytesTypeStr | StringTypeStr | DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes | str | int, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[BytesTypeStr | StringTypeStr | BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes | str | bool, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[BytesTypeStr | DecodesToIntTypeStr | BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes | int | bool, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[StringTypeStr | DecodesToIntTypeStr | BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[str | int | bool, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[
            AddressTypeStr | BytesTypeStr | StringTypeStr | DecodesToIntTypeStr
        ],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress | bytes | str | int, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[AddressTypeStr | BytesTypeStr | StringTypeStr | BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress | bytes | str | bool, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[
            AddressTypeStr | BytesTypeStr | DecodesToIntTypeStr | BoolTypeStr
        ],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress | bytes | int | bool, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[
            AddressTypeStr | StringTypeStr | DecodesToIntTypeStr | BoolTypeStr
        ],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[AddressTypeStr | str | int | bool, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[
            BytesTypeStr | StringTypeStr | DecodesToIntTypeStr | BoolTypeStr
        ],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[bytes | str | int | bool, ...]: ...
    @overload
    def decode(
        self,
        types: Iterable[
            AddressTypeStr
            | BytesTypeStr
            | StringTypeStr
            | DecodesToIntTypeStr
            | BoolTypeStr
        ],
        data: Decodable,
        strict: bool = True,
    ) -> tuple[HexAddress | bytes | str | int | bool, ...]: ...
    @overload
    def decode(
        self, types: Iterable[TypeStr], data: Decodable, strict: bool = True
    ) -> tuple[Any, ...]: ...

class ABICodec(ABIEncoder, ABIDecoder): ...



import pytest

from faster_eth_abi.utils.validation import (
    validate_bytes_param,
    validate_list_like_param,
)


@pytest.mark.parametrize(
    "bytes_value",
    (
        b"",
        b"\x00",
        b"david attenborough",
        b"\xff" * 32,
        bytearray(b"\x01\x02\x03\x04\xff"),
    ),
)
def test_validate_bytes_param(bytes_value):
    validate_bytes_param(bytes_value, "test_param")


@pytest.mark.parametrize(
    "non_bytes_value",
    (
        "",
        123,
        [b"david attenborough"],
        {b"david attenborough"},
        (b"david attenborough",),
    ),
)
def test_validate_bytes_param_raises_for_non_bytes_types(non_bytes_value):
    param_name = "test_param"
    with pytest.raises(
        TypeError,
        match=f"The `{param_name}` value must be of bytes type. "
        f"Got {type(non_bytes_value)}",
    ):
        validate_bytes_param(non_bytes_value, param_name)


@pytest.mark.parametrize(
    "list_like_value", ([1, 2], (1, 2), ["1", 2], ("1", 2), [], tuple())
)
def test_validate_list_like_param(list_like_value):
    validate_list_like_param(list_like_value, "test_param")


@pytest.mark.parametrize(
    "non_list_like_value",
    (
        "",
        123,
        b"",
        b"\xff",
        b"david attenborough",
        bytearray(b"\x01\xff"),
        {"key": "val"},
        {1, 2},
    ),
)
def test_validate_list_like_param_taises_for_non_list_like_types(non_list_like_value):
    param_name = "test_param"
    with pytest.raises(
        TypeError,
        match=f"The `{param_name}` value type must be one of list or tuple. "
        f"Got {type(non_list_like_value)}",
    ):
        validate_list_like_param(non_list_like_value, param_name)



# mypy: disable-error-code="overload-overlap"
"""
API for ABI encoding and decoding.

Defines the main encoder and decoder classes, providing methods for binary serialization
and deserialization of values according to ABI type specifications.
"""

from typing import (
    Any,
    Iterable,
    Tuple,
    Union,
    overload,
)

from eth_typing import (
    HexAddress,
)
from eth_typing.abi import (
    Decodable,
    TypeStr,
)

from faster_eth_abi._codec import (
    decode_c,
    encode_c,
)
from faster_eth_abi.decoding import (
    ContextFramesBytesIO,
)
from faster_eth_abi.exceptions import (
    EncodingError,
    MultipleEntriesFound,
)
from faster_eth_abi.registry import (
    ABIRegistry,
)
from faster_eth_abi.typing import (
    AddressTypeStr,
    BoolTypeStr,
    BytesTypeStr,
    IntTypeStr,
    StringTypeStr,
    UintTypeStr,
)

DecodesToIntTypeStr = Union[UintTypeStr, IntTypeStr]


class BaseABICoder:
    """
    Base class for porcelain coding APIs.  These are classes which wrap
    instances of :class:`~faster_eth_abi.registry.ABIRegistry` to provide last-mile
    coding functionality.
    """

    def __init__(self, registry: ABIRegistry):
        """
        Constructor.

        :param registry: The registry providing the encoders to be used when
            encoding values.
        """
        self._registry = registry


class ABIEncoder(BaseABICoder):
    """
    Wraps a registry to provide last-mile encoding functionality.
    """

    def encode(self, types: Iterable[TypeStr], args: Iterable[Any]) -> bytes:
        """
        Encodes the python values in ``args`` as a sequence of binary values of
        the ABI types in ``types`` via the head-tail mechanism.

        :param types: A list or tuple of string representations of the ABI types
            that will be used for encoding e.g.  ``('uint256', 'bytes[]',
            '(int,int)')``
        :param args: A list or tuple of python values to be encoded.

        :returns: The head-tail encoded binary representation of the python
            values in ``args`` as values of the ABI types in ``types``.
        """
        return encode_c(self, types, args)

    def is_encodable(self, typ: TypeStr, arg: Any) -> bool:
        """
        Determines if the python value ``arg`` is encodable as a value of the
        ABI type ``typ``.

        :param typ: A string representation for the ABI type against which the
            python value ``arg`` will be checked e.g. ``'uint256'``,
            ``'bytes[]'``, ``'(int,int)'``, etc.
        :param arg: The python value whose encodability should be checked.

        :returns: ``True`` if ``arg`` is encodable as a value of the ABI type
            ``typ``.  Otherwise, ``False``.
        """
        try:
            encoder = self._registry.get_encoder(typ)
        except MultipleEntriesFound:
            raise
        except Exception:
            return False

        validate = getattr(encoder, "validate_value", encoder)
        try:
            validate(arg)
        except EncodingError:
            return False

        return True

    def is_encodable_type(self, typ: TypeStr) -> bool:
        """
        Returns ``True`` if values for the ABI type ``typ`` can be encoded by
        this codec.

        :param typ: A string representation for the ABI type that will be
            checked for encodability e.g. ``'uint256'``, ``'bytes[]'``,
            ``'(int,int)'``, etc.

        :returns: ``True`` if values for ``typ`` can be encoded by this codec.
            Otherwise, ``False``.
        """
        return self._registry.has_encoder(typ)


class ABIDecoder(BaseABICoder):
    """
    Wraps a registry to provide last-mile decoding functionality.
    """

    stream_class = ContextFramesBytesIO

    # raw tuple types, same type

    # len == 1

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any]:
        ...

    # len == 2

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, Any]:
        ...

    # len == 3
    # okay get ready for some ugly overloads
    # We will probably not implement lengths > 3

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, AddressTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, HexAddress, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, AddressTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, HexAddress, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, AddressTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, HexAddress, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, AddressTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, HexAddress, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, AddressTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, HexAddress, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, AddressTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, HexAddress, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, BytesTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, bytes, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, BytesTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, bytes, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, BytesTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, bytes, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, BytesTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, bytes, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, BytesTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, bytes, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, BytesTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, bytes, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, StringTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, str, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, StringTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, str, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, StringTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, str, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, StringTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, str, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, StringTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, str, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, StringTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, str, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, DecodesToIntTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, int, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, DecodesToIntTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, int, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, DecodesToIntTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, int, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, DecodesToIntTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, int, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, DecodesToIntTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, int, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, DecodesToIntTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, int, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, BoolTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, bool, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, BoolTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, bool, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, BoolTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, bool, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, BoolTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, bool, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, BoolTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, bool, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, BoolTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, bool, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, TypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, Any, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, TypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, Any, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, TypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, Any, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, TypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, Any, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, TypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, Any, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[AddressTypeStr, TypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, Any, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, AddressTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, HexAddress, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, AddressTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, HexAddress, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, AddressTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, HexAddress, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, AddressTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, HexAddress, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, AddressTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, HexAddress, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, AddressTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, HexAddress, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, BytesTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, bytes, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, BytesTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, bytes, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, BytesTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, bytes, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, BytesTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, bytes, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, BytesTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, bytes, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, BytesTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, bytes, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, StringTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, str, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, StringTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, str, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, StringTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, str, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, StringTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, str, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, StringTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, str, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, StringTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, str, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, DecodesToIntTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, int, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, DecodesToIntTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, int, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, DecodesToIntTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, int, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, DecodesToIntTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, int, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, DecodesToIntTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, int, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, DecodesToIntTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, int, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, BoolTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, bool, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, BoolTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, bool, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, BoolTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, bool, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, BoolTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, bool, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, BoolTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, bool, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, BoolTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, bool, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, TypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, Any, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, TypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, Any, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, TypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, Any, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, TypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, Any, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, TypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, Any, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BytesTypeStr, TypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, Any, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, AddressTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, HexAddress, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, AddressTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, HexAddress, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, AddressTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, HexAddress, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, AddressTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, HexAddress, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, AddressTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, HexAddress, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, AddressTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, HexAddress, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, BytesTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, bytes, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, BytesTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, bytes, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, BytesTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, bytes, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, BytesTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, bytes, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, BytesTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, bytes, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, BytesTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, bytes, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, StringTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, str, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, StringTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, str, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, StringTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, str, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, StringTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, str, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, StringTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, str, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, StringTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, str, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, DecodesToIntTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, int, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, DecodesToIntTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, int, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, DecodesToIntTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, int, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, DecodesToIntTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, int, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, DecodesToIntTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, int, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, DecodesToIntTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, int, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, BoolTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, bool, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, BoolTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, bool, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, BoolTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, bool, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, BoolTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, bool, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, BoolTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, bool, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, BoolTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, bool, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, TypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, Any, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, TypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, Any, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, TypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, Any, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, TypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, Any, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, TypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, Any, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[StringTypeStr, TypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, Any, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, AddressTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, HexAddress, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, AddressTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, HexAddress, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, AddressTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, HexAddress, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, AddressTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, HexAddress, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, AddressTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, HexAddress, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, AddressTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, HexAddress, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, BytesTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, bytes, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, BytesTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, bytes, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, BytesTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, bytes, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, BytesTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, bytes, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, BytesTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, bytes, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, BytesTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, bytes, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, StringTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, str, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, StringTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, str, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, StringTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, str, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, StringTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, str, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, StringTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, str, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, StringTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, str, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, DecodesToIntTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, int, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, DecodesToIntTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, int, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, DecodesToIntTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, int, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, DecodesToIntTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, int, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, DecodesToIntTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, int, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, DecodesToIntTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, int, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, BoolTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, bool, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, BoolTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, bool, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, BoolTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, bool, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, BoolTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, bool, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, BoolTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, bool, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, BoolTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, bool, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, TypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, Any, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, TypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, Any, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, TypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, Any, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, TypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, Any, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, TypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, Any, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[DecodesToIntTypeStr, TypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, Any, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, AddressTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, HexAddress, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, AddressTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, HexAddress, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, AddressTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, HexAddress, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, AddressTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, HexAddress, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, AddressTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, HexAddress, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, AddressTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, HexAddress, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, BytesTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, bytes, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, BytesTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, bytes, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, BytesTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, bytes, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, BytesTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, bytes, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, BytesTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, bytes, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, BytesTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, bytes, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, StringTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, str, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, StringTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, str, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, StringTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, str, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, StringTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, str, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, StringTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, str, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, StringTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, str, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, DecodesToIntTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, int, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, DecodesToIntTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, int, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, DecodesToIntTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, int, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, DecodesToIntTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, int, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, DecodesToIntTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, int, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, DecodesToIntTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, int, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, BoolTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, bool, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, BoolTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, bool, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, BoolTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, bool, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, BoolTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, bool, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, BoolTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, bool, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, BoolTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, bool, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, TypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, Any, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, TypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, Any, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, TypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, Any, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, TypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, Any, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, TypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, Any, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[BoolTypeStr, TypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, Any, Any]:
        ...

    # --- Fallbacks for len == 3 ---

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, AddressTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, HexAddress, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, AddressTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, HexAddress, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, AddressTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, HexAddress, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, AddressTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, HexAddress, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, AddressTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, HexAddress, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, AddressTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, HexAddress, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, BytesTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, bytes, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, BytesTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, bytes, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, BytesTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, bytes, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, BytesTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, bytes, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, BytesTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, bytes, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, BytesTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, bytes, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, StringTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, str, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, StringTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, str, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, StringTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, str, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, StringTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, str, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, StringTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, str, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, StringTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, str, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, DecodesToIntTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, int, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, DecodesToIntTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, int, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, DecodesToIntTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, int, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, DecodesToIntTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, int, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, DecodesToIntTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, int, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, DecodesToIntTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, int, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, BoolTypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, bool, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, BoolTypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, bool, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, BoolTypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, bool, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, BoolTypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, bool, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, BoolTypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, bool, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, BoolTypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, bool, Any]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, TypeStr, AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, Any, HexAddress]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, TypeStr, BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, Any, bytes]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, TypeStr, StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, Any, str]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, TypeStr, DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, Any, int]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, TypeStr, BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, Any, bool]:
        ...

    @overload
    def decode(
        self,
        types: Tuple[TypeStr, TypeStr, TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, Any, Any]:
        ...

    # non-tuple types input

    @overload
    def decode(
        self,
        types: Iterable[AddressTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[HexAddress, ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[BytesTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bytes, ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[StringTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[str, ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[DecodesToIntTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[int, ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[BoolTypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[bool, ...]:
        ...

    # fallback to union types, still better than Any

    @overload
    def decode(
        self,
        types: Iterable[Union[AddressTypeStr, BytesTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[HexAddress, bytes], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[AddressTypeStr, StringTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[HexAddress, str], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[AddressTypeStr, DecodesToIntTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[HexAddress, int], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[AddressTypeStr, BoolTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[HexAddress, bool], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[BytesTypeStr, StringTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[bytes, str], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[BytesTypeStr, DecodesToIntTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[bytes, int], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[BytesTypeStr, BoolTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[bytes, bool], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[StringTypeStr, DecodesToIntTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[str, int], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[StringTypeStr, BoolTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[str, bool], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[DecodesToIntTypeStr, BoolTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[int, bool], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[AddressTypeStr, BytesTypeStr, StringTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[HexAddress, bytes, str], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[AddressTypeStr, BytesTypeStr, DecodesToIntTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[HexAddress, bytes, int], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[AddressTypeStr, BytesTypeStr, BoolTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[HexAddress, bytes, bool], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[AddressTypeStr, StringTypeStr, DecodesToIntTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[HexAddress, str, int], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[AddressTypeStr, StringTypeStr, BoolTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[HexAddress, str, bool], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[AddressTypeStr, DecodesToIntTypeStr, BoolTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[HexAddress, int, bool], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[BytesTypeStr, StringTypeStr, DecodesToIntTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[bytes, str, int], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[BytesTypeStr, StringTypeStr, BoolTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[bytes, str, bool], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[BytesTypeStr, DecodesToIntTypeStr, BoolTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[bytes, int, bool], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[Union[StringTypeStr, DecodesToIntTypeStr, BoolTypeStr]],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[str, int, bool], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[
            Union[AddressTypeStr, BytesTypeStr, StringTypeStr, DecodesToIntTypeStr]
        ],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[HexAddress, bytes, str, int], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[
            Union[AddressTypeStr, BytesTypeStr, StringTypeStr, BoolTypeStr]
        ],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[HexAddress, bytes, str, bool], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[
            Union[AddressTypeStr, BytesTypeStr, DecodesToIntTypeStr, BoolTypeStr]
        ],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[HexAddress, bytes, int, bool], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[
            Union[AddressTypeStr, StringTypeStr, DecodesToIntTypeStr, BoolTypeStr]
        ],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[AddressTypeStr, str, int, bool], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[
            Union[BytesTypeStr, StringTypeStr, DecodesToIntTypeStr, BoolTypeStr]
        ],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[bytes, str, int, bool], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[
            Union[
                AddressTypeStr,
                BytesTypeStr,
                StringTypeStr,
                DecodesToIntTypeStr,
                BoolTypeStr,
            ]
        ],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Union[HexAddress, bytes, str, int, bool], ...]:
        ...

    @overload
    def decode(
        self,
        types: Iterable[TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, ...]:
        ...

    def decode(
        self,
        types: Iterable[TypeStr],
        data: Decodable,
        strict: bool = True,
    ) -> Tuple[Any, ...]:
        """
        Decodes the binary value ``data`` as a sequence of values of the ABI types
        in ``types`` via the head-tail mechanism into a tuple of equivalent python
        values.

        :param types: A list or tuple of string representations of the ABI types that
            will be used for decoding e.g. ``('uint256', 'bytes[]', '(int,int)')``
        :param data: The binary value to be decoded.
        :param strict: If ``False``, dynamic-type decoders will ignore validations such
            as making sure the data is padded to a multiple of 32 bytes or checking that
            padding bytes are zero / empty. ``False`` is how the Solidity ABI decoder
            currently works. However, ``True`` is the default for the faster-eth-abi
            library.

        :returns: A tuple of equivalent python values for the ABI values
            represented in ``data``.
        """
        return decode_c(self, types, data, strict)


class ABICodec(ABIEncoder, ABIDecoder):
    pass



from typing import Any

def abbr(value: Any, limit: int = 79) -> str:
    """
    Converts a value into its string representation and abbreviates that
    representation based on the given length `limit` if necessary.
    """



from .decoding import (
    BaseArrayDecoder as BaseArrayDecoder,
    DynamicArrayDecoder as DynamicArrayDecoder,
    FixedByteSizeDecoder as FixedByteSizeDecoder,
    HeadTailDecoder as HeadTailDecoder,
    SignedIntegerDecoder as SignedIntegerDecoder,
    SizedArrayDecoder as SizedArrayDecoder,
    TupleDecoder as TupleDecoder,
)
from faster_eth_abi.exceptions import (
    InsufficientDataBytes as InsufficientDataBytes,
    InvalidPointer as InvalidPointer,
    NonEmptyPaddingBytes as NonEmptyPaddingBytes,
)
from faster_eth_abi.io import (
    BytesIO as BytesIO,
    ContextFramesBytesIO as ContextFramesBytesIO,
)
from faster_eth_abi.typing import T as T

def decode_uint_256(stream: ContextFramesBytesIO) -> int:
    """
    A faster version of :func:`~decoding.decode_uint_256` in decoding.py.

    It recreates the logic from the UnsignedIntegerDecoder, but we can
    skip a lot because we know the value of many vars.
    """

def get_value_byte_size(decoder: FixedByteSizeDecoder) -> int: ...
def decode_head_tail(self, stream: ContextFramesBytesIO) -> T: ...
def decode_tuple(self, stream: ContextFramesBytesIO) -> tuple[T, ...]: ...
def validate_pointers_tuple(self, stream: ContextFramesBytesIO) -> None:
    """
    Verify that all pointers point to a valid location in the stream.
    """

def validate_pointers_array(
    self, stream: ContextFramesBytesIO, array_size: int
) -> None:
    """
    Verify that all pointers point to a valid location in the stream.
    """

def decode_sized_array(self, stream: ContextFramesBytesIO) -> tuple[T, ...]: ...
def decode_dynamic_array(self, stream: ContextFramesBytesIO) -> tuple[T, ...]: ...
def read_fixed_byte_size_data_from_stream(self, stream: BytesIO) -> bytes: ...
def split_data_and_padding_fixed_byte_size(
    self, raw_data: bytes
) -> tuple[bytes, bytes]: ...
def validate_padding_bytes_fixed_byte_size(
    self, value: T, padding_bytes: bytes
) -> None: ...
def get_expected_padding_bytes(self, chunk: bytes) -> bytes: ...
def validate_padding_bytes_signed_integer(
    self, value: int, padding_bytes: bytes
) -> None: ...
def decoder_fn_boolean(data: bytes) -> bool: ...



from _typeshed import Incomplete
from eth_typing.abi import TypeStr as TypeStr
from faster_eth_abi._grammar import (
    ABIType as ABIType,
    BasicType as BasicType,
    TupleType as TupleType,
    normalize as normalize,
)
from faster_eth_abi.grammar import parse as parse
from faster_eth_abi.registry import (
    BaseEquals as BaseEquals,
    BaseRegistry as BaseRegistry,
    Lookup as Lookup,
    PredicateMapping as PredicateMapping,
    has_arrlist as has_arrlist,
    is_base_tuple as is_base_tuple,
)
from faster_eth_abi.utils.numeric import scale_places as scale_places
from hypothesis import strategies as st
from typing import Final

StrategyFactory: Incomplete
StrategyRegistration: Incomplete

class StrategyRegistry(BaseRegistry):
    def __init__(self) -> None: ...
    def register_strategy(
        self,
        lookup: Lookup,
        registration: StrategyRegistration,
        label: str | None = None,
    ) -> None: ...
    def unregister_strategy(self, lookup_or_label: Lookup) -> None: ...
    def get_strategy(self, type_str: TypeStr) -> st.SearchStrategy:
        """
        Returns a hypothesis strategy for the given ABI type.

        :param type_str: The canonical string representation of the ABI type
            for which a hypothesis strategy should be returned.

        :returns: A hypothesis strategy for generating Python values that are
            encodable as values of the given ABI type.
        """

def get_uint_strategy(
    abi_type: BasicType, registry: StrategyRegistry
) -> st.SearchStrategy: ...
def get_int_strategy(
    abi_type: BasicType, registry: StrategyRegistry
) -> st.SearchStrategy: ...

address_strategy: Final[Incomplete]
bool_strategy: Final[Incomplete]

def get_ufixed_strategy(
    abi_type: BasicType, registry: StrategyRegistry
) -> st.SearchStrategy: ...
def get_fixed_strategy(
    abi_type: BasicType, registry: StrategyRegistry
) -> st.SearchStrategy: ...
def get_bytes_strategy(
    abi_type: BasicType, registry: StrategyRegistry
) -> st.SearchStrategy: ...

bytes_strategy: Final[Incomplete]
string_strategy: Final[Incomplete]

def get_array_strategy(
    abi_type: ABIType, registry: StrategyRegistry
) -> st.SearchStrategy: ...
def get_tuple_strategy(
    abi_type: TupleType, registry: StrategyRegistry
) -> st.SearchStrategy: ...

strategy_registry: Final[Incomplete]
get_abi_strategy: Final[Incomplete]



"""Helpers for packed ABI encoding.

Defines functions and registry for packed encoding and encodability checks.
"""
from typing import (
    Final,
)

from .codec import (
    ABIEncoder,
)
from .registry import (
    registry_packed,
)

default_encoder_packed: Final = ABIEncoder(registry_packed)

encode_packed: Final = default_encoder_packed.encode
is_encodable_packed: Final = default_encoder_packed.is_encodable
